@@ -14,7 +14,7 @@
 
 	You should have received a copy of the GNU General Public License
 	along with this program; if not, write to the Free Software
-	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+	Foundation Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
 
 #define __NO_VERSION__
@@ -37,7 +37,6 @@
 #include <asm/pgtable.h>
 #include <asm/page.h>
 #include <linux/sched.h>
-#include <asm/segment.h>
 #include <linux/wait.h>
 
 #include <linux/version.h>
