@@ -75,6 +75,20 @@ static int em8300_getsda(void *data)
 	return readl(&em->mem[em->i2c_pin_reg]) & (p->data << 8);
 }
 
+static int em8300_i2c_lock_client(struct i2c_client *client)
+{
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,54)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
+	if (!try_module_get(client->driver->driver.owner)) {
+#else
+	if (!try_module_get(client->driver->owner)) {
+#endif
+		printk(KERN_ERR "em8300_i2c: Unable to lock client module\n");
+		return -ENODEV;
+	}
+#endif
+	return 0;
+}
 
 static int em8300_i2c_reg(struct i2c_client *client)
 {
@@ -82,6 +96,9 @@ static int em8300_i2c_reg(struct i2c_client *client)
 
 	switch (client->driver->id) {
 	case I2C_DRIVERID_ADV717X:
+		if (em8300_i2c_lock_client(client)) {
+			return -ENODEV;
+		}
 		if (!strncmp(client->name, "ADV7175", 7)) {
 			em->encoder_type = ENCODER_ADV7175;
 		}
@@ -91,6 +108,9 @@ static int em8300_i2c_reg(struct i2c_client *client)
 		em->encoder = client;
 		break;
 	case I2C_DRIVERID_BT865:
+		if (em8300_i2c_lock_client(client)) {
+			return -ENODEV;
+		}
 		em->encoder_type = ENCODER_BT865;
 		em->encoder = client;
 		break;
@@ -111,10 +131,15 @@ static int em8300_i2c_unreg(struct i2c_client *client)
 
 	switch (client->driver->id) {
 	case I2C_DRIVERID_ADV717X:
+	case I2C_DRIVERID_BT865:
 		em->encoder = NULL;
-		break;
-	case  I2C_DRIVERID_BT865:
-		em->encoder = NULL;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,54)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
+		module_put(client->driver->driver.owner);
+#else
+		module_put(client->driver->owner);
+#endif
+#endif
 		break;
 	}
 
@@ -159,7 +184,6 @@ int em8300_i2c_init(struct em8300_s *em)
 	em->i2c_data_1.setscl = &em8300_setscl;
 	em->i2c_data_1.getsda = &em8300_getsda;
 	em->i2c_data_1.getscl = &em8300_getscl;
-	em->i2c_data_1.mdelay = 10;
 	em->i2c_data_1.udelay = 10;
 	em->i2c_data_1.timeout = 100;
 
@@ -193,7 +217,6 @@ int em8300_i2c_init(struct em8300_s *em)
 	em->i2c_data_2.setscl = &em8300_setscl;
 	em->i2c_data_2.getsda = &em8300_getsda;
 	em->i2c_data_2.getscl = &em8300_getscl;
-	em->i2c_data_2.mdelay = 10;
 	em->i2c_data_2.udelay = 10;
 	em->i2c_data_2.timeout = 100;
 
