@@ -30,6 +30,16 @@
 #include "em8300_registration.h"
 #include "em8300_compat24.h"
 
+typedef enum {
+	AUDIO_DRIVER_NONE,
+	AUDIO_DRIVER_OSSLIKE,
+	AUDIO_DRIVER_OSS,
+	AUDIO_DRIVER_ALSA,
+	AUDIO_DRIVER_MAX
+} audio_driver_t;
+
+extern audio_driver_t audio_driver_nr[EM8300_MAX];
+
 int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 {
 	em8300_register_t reg;
@@ -65,6 +75,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		return em8300_ioctl_init(em, (em8300_microcode_t *) arg);
 
 	case _IOC_NR(EM8300_IOCTL_WRITEREG):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -80,6 +92,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_READREG):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -98,6 +112,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_GETSTATUS):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -106,6 +122,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_VBI):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -126,6 +144,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		return 0;
 
 	case _IOC_NR(EM8300_IOCTL_GETBCS):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -146,6 +166,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_SET_VIDEOMODE):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -162,6 +184,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_SET_PLAYMODE):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -173,6 +197,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_SET_ASPECTRATIO):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -187,22 +213,29 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 				return -EFAULT;
 		}
 		break;
-
 	case _IOC_NR(EM8300_IOCTL_GET_AUDIOMODE):
-		if (!em->ucodeloaded) {
-			return -ENOTTY;
-		}
+		if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+		    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS)) {
+			em8300_require_ucode(em);
 
-		if (_IOC_DIR(cmd) & _IOC_WRITE) {
-			get_user(val, (int *) arg);
-			em8300_ioctl_setaudiomode(em, val);
-		}
-		if (_IOC_DIR(cmd) & _IOC_READ) {
-			em8300_ioctl_getaudiomode(em, arg);
-		}
-		break;
+			if (!em->ucodeloaded) {
+				return -ENOTTY;
+			}
 
+			if (_IOC_DIR(cmd) & _IOC_WRITE) {
+				get_user(val, (int *) arg);
+				em8300_ioctl_setaudiomode(em, val);
+			}
+			if (_IOC_DIR(cmd) & _IOC_READ) {
+				em8300_ioctl_getaudiomode(em, arg);
+			}
+			break;
+		}
+		else
+			return -EINVAL;
 	case _IOC_NR(EM8300_IOCTL_SET_SPUMODE):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -219,6 +252,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_OVERLAY_SETMODE):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -232,6 +267,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_OVERLAY_SIGNALMODE):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -245,6 +282,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_OVERLAY_SETWINDOW):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -263,6 +302,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_OVERLAY_SETSCREEN):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -281,6 +322,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 	break;
 
 	case _IOC_NR(EM8300_IOCTL_OVERLAY_CALIBRATE):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -300,6 +343,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 	break;
 
 	case _IOC_NR(EM8300_IOCTL_OVERLAY_GET_ATTRIBUTE):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -317,6 +362,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 		break;
 
 	case _IOC_NR(EM8300_IOCTL_SCR_GET):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -346,6 +393,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 	break;
 
 	case _IOC_NR(EM8300_IOCTL_SCR_GETSPEED):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -368,6 +417,8 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 	break;
 
 	case _IOC_NR(EM8300_IOCTL_FLUSH):
+		em8300_require_ucode(em);
+
 		if (!em->ucodeloaded) {
 			return -ENOTTY;
 		}
@@ -382,7 +433,11 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 			case EM8300_SUBDEVICE_VIDEO:
 				return em8300_video_flush(em);
 			case EM8300_SUBDEVICE_AUDIO:
-				return em8300_audio_flush(em);
+				if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+				    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+					return em8300_audio_flush(em);
+				else
+					return -EINVAL;
 			case EM8300_SUBDEVICE_SUBPICTURE:
 				return -ENOSYS;
 			default:
@@ -401,15 +456,26 @@ int em8300_control_ioctl(struct em8300_s *em, int cmd, unsigned long arg)
 int em8300_ioctl_init(struct em8300_s *em, em8300_microcode_t *useruc)
 {
 	em8300_microcode_t uc;
+	unsigned char *ucode;
 	int ret;
 
 	if (copy_from_user(&uc, useruc, sizeof(em8300_microcode_t)))
 		return -EFAULT;
 
-	if ((ret = em8300_ucode_upload(em, uc.ucode, uc.ucode_size))) {
-		return ret;
+	ucode = kmalloc(uc.ucode_size, GFP_KERNEL);
+	if (!ucode) {
+		return -ENOMEM;
 	}
 
+	if (copy_from_user(ucode, uc.ucode, uc.ucode_size)) {
+		kfree(ucode);
+		return -EFAULT;
+	}
+
+	em8300_ucode_upload(em, ucode, uc.ucode_size);
+
+	kfree(ucode);
+
 	em8300_dicom_init(em);
 
 	if ((ret = em8300_video_setup(em))) {
@@ -419,9 +485,11 @@ int em8300_ioctl_init(struct em8300_s *em, em8300_microcode_t *useruc)
 	if (em->mvfifo) {
 		em8300_fifo_free(em->mvfifo);
 	}
-	if (em->mafifo) {
-		em8300_fifo_free(em->mafifo);
-	}
+	if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+	    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+		if (em->mafifo) {
+			em8300_fifo_free(em->mafifo);
+		}
 	if (em->spfifo) {
 		em8300_fifo_free(em->spfifo);
 	}
@@ -430,28 +498,36 @@ int em8300_ioctl_init(struct em8300_s *em, em8300_microcode_t *useruc)
 		return -ENOMEM;
 	}
 
-	if (!(em->mafifo = em8300_fifo_alloc())) {
-		return -ENOMEM;
-	}
+	if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+	    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+		if (!(em->mafifo = em8300_fifo_alloc())) {
+			return -ENOMEM;
+		}
 
 	if (!(em->spfifo = em8300_fifo_alloc())) {
 		return -ENOMEM;
 	}
 
 	em8300_fifo_init(em,em->mvfifo, MV_PCIStart, MV_PCIWrPtr, MV_PCIRdPtr, MV_PCISize, 0x900, FIFOTYPE_VIDEO);
-	em8300_fifo_init(em,em->mafifo, MA_PCIStart, MA_PCIWrPtr, MA_PCIRdPtr, MA_PCISize, 0x1000, FIFOTYPE_AUDIO);
+	if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+	    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+		em8300_fifo_init(em,em->mafifo, MA_PCIStart, MA_PCIWrPtr, MA_PCIRdPtr, MA_PCISize, 0x1000, FIFOTYPE_AUDIO);
 	//	em8300_fifo_init(em,em->spfifo, SP_PCIStart, SP_PCIWrPtr, SP_PCIRdPtr, SP_PCISize, 0x1000, FIFOTYPE_VIDEO);
 	em8300_fifo_init(em,em->spfifo, SP_PCIStart, SP_PCIWrPtr, SP_PCIRdPtr, SP_PCISize, 0x800, FIFOTYPE_VIDEO);
 	em8300_spu_init(em);
 
-	if ((ret = em8300_audio_setup(em))) {
-		return ret;
-	}
+	if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+	    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+		if ((ret = em8300_audio_setup(em))) {
+			return ret;
+		}
 
 	em8300_ioctl_enable_videoout(em, 1);
 
-	if (em->ucodeloaded == 0)
+#if ! ( defined(CONFIG_FW_LOADER) || defined(CONFIG_FW_LOADER_MODULE) )
+	if (!em->ucodeloaded)
 		em8300_enable_card(em);
+#endif
 
 	em->ucodeloaded = 1;
 
@@ -469,7 +545,9 @@ int em8300_ioctl_getstatus(struct em8300_s *em, char *usermsg)
 	char spfstatus[128];
 
 	em8300_fifo_statusmsg(em->mvfifo, mvfstatus);
-	em8300_fifo_statusmsg(em->mafifo, mafstatus);
+	if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+	    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+		em8300_fifo_statusmsg(em->mafifo, mafstatus);
 	em8300_fifo_statusmsg(em->spfifo, spfstatus);
 
 	frames = (read_ucregister(MV_FrameCntHi) << 16) | read_ucregister(MV_FrameCntLo);
