@@ -15,11 +15,13 @@
 
 	You should have received a copy of the GNU General Public License
 	along with this program; if not, write to the Free Software
-	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+	Foundation Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
-#include <linux/config.h>
 #include <linux/version.h>
 #include <linux/module.h>
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
+#include <linux/moduleparam.h>
+#endif
 #include <linux/delay.h>
 #include <linux/init.h>
 #include <linux/errno.h>
@@ -35,12 +37,10 @@
 #include <linux/string.h>
 #include <linux/time.h>
 #include <linux/poll.h>
-#include <linux/devfs_fs_kernel.h>
 #include <asm/io.h>
 #include <asm/pgtable.h>
 #include <asm/page.h>
 #include <linux/sched.h>
-#include <asm/segment.h>
 #ifdef CONFIG_MTRR
 #include <asm/mtrr.h>
 #endif
@@ -73,11 +73,18 @@ MODULE_LICENSE("GPL");
 #if EM8300_MAJOR != 0
 MODULE_ALIAS_CHARDEV_MAJOR(EM8300_MAJOR);
 #endif
+#ifdef MODULE_VERSION
+MODULE_VERSION(EM8300_VERSION);
+#endif
 
 EXPORT_NO_SYMBOLS;
 
-static unsigned int use_bt865[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
+static int use_bt865[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(use_bt865, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(use_bt865, bool, NULL, 0444);
+#endif
 MODULE_PARM_DESC(use_bt865, "Set this to 1 if you have a bt865. It changes some internal register values. Defaults to 0.");
 
 /*
@@ -88,7 +95,11 @@ int dicom_other_pal[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 1 };
 #else
 int dicom_other_pal[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 #endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(dicom_other_pal, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(dicom_other_pal, bool, NULL, 0444);
+#endif
 MODULE_PARM_DESC(dicom_other_pal, "If this is set, then some internal register values are swapped for PAL and NTSC. Defaults to 1.");
 
 #ifdef CONFIG_EM8300_DICOMFIX
@@ -96,7 +107,11 @@ int dicom_fix[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 1 };
 #else
 int dicom_fix[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 #endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(dicom_fix, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(dicom_fix, bool, NULL, 0444);
+#endif
 MODULE_PARM_DESC(dicom_fix, "If this is set then some internal register values are changed. Fixes green screen problems for some. Defaults to 1.");
 
 #ifdef CONFIG_EM8300_DICOMCTRL
@@ -104,7 +119,11 @@ int dicom_control[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 1 };
 #else
 int dicom_control[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 #endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(dicom_control, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(dicom_control, bool, NULL, 0444);
+#endif
 MODULE_PARM_DESC(dicom_control, "If this is set then some internal register values are changed. Fixes green screen problems for some. Defaults to 1.");
 
 #ifdef CONFIG_EM8300_UCODETIMEOUT
@@ -112,7 +131,11 @@ int bt865_ucode_timeout[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 1 };
 #else
 int bt865_ucode_timeout[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 #endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(bt865_ucode_timeout, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(bt865_ucode_timeout, bool, NULL, 0444);
+#endif
 MODULE_PARM_DESC(bt865_ucode_timeout, "Set this to 1 if you have a bt865 and get timeouts when uploading the microcode. Defaults to 0.");
 
 #ifdef CONFIG_EM8300_LOOPBACK
@@ -120,11 +143,19 @@ int activate_loopback[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 1 };
 #else
 int activate_loopback[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = 0 };
 #endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(activate_loopback, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(activate_loopback, bool, NULL, 0444);
+#endif
 MODULE_PARM_DESC(activate_loopback, "If you lose video after loading the modules or uploading the microcode set this to 1. Defaults to 0.");
 
 int major = EM8300_MAJOR;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(major, "i");
+#else
+module_param(major, int, 0444);
+#endif
 MODULE_PARM_DESC(major, "Major number used for the devices. "
 		 "0 means automatically assigned. "
 		 "Defaults to " __MODULE_STRING(EM8300_MAJOR) ".");
@@ -133,14 +164,86 @@ static int em8300_cards,clients;
 
 static struct em8300_s em8300[EM8300_MAX];
 
+typedef enum {
+	AUDIO_DRIVER_NONE,
+	AUDIO_DRIVER_OSSLIKE,
+	AUDIO_DRIVER_OSS,
+	AUDIO_DRIVER_ALSA,
+	AUDIO_DRIVER_MAX
+} audio_driver_t;
+
+static const char * const audio_driver_name[] = {
+	[ AUDIO_DRIVER_NONE ] = "none",
+	[ AUDIO_DRIVER_OSSLIKE ] = "osslike",
+	[ AUDIO_DRIVER_OSS ] = "oss",
+	[ AUDIO_DRIVER_ALSA ] = "alsa",
+};
+
+#if defined(CONFIG_SND) || defined(CONFIG_SND_MODULE)
+audio_driver_t audio_driver_nr[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = AUDIO_DRIVER_ALSA };
+#elif defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
+audio_driver_t audio_driver_nr[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = AUDIO_DRIVER_OSS };
+#else
+audio_driver_t audio_driver_nr[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = AUDIO_DRIVER_OSSLIKE };
+#endif
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
+static char *audio_driver[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = NULL };
+MODULE_PARM(audio_driver, "1-" __MODULE_STRING(EM8300_MAX) "s");
+#else
+static int param_set_audio_driver_t(const char *val, struct kernel_param *kp)
+{
+	if (val) {
+		int i;
+		for (i=0; i < AUDIO_DRIVER_MAX; i++)
+			if (strcmp(val, audio_driver_name[i]) == 0) {
+				*(audio_driver_t *)kp->arg = i;
+				return 0;
+			}
+	}
+	printk(KERN_ERR "%s: audio_driver parameter expected\n",
+	       kp->name);
+	return -EINVAL;
+}
+
+static int param_get_audio_driver_t(char *buffer, struct kernel_param *kp)
+{
+	return sprintf(buffer, "%s", audio_driver_name[*(audio_driver_t *)kp->arg]);
+}
+
+module_param_array_named(audio_driver, audio_driver_nr, audio_driver_t, NULL, 0444);
+#endif
+MODULE_PARM_DESC(audio_driver, "The audio driver to use (none, osslike, oss, or alsa).");
+
 #if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
 int dsp_num[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = -1 };
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 MODULE_PARM(dsp_num, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(dsp_num, int, NULL, 0444);
+#endif
 MODULE_PARM_DESC(dsp_num, "The /dev/dsp number to assign to the card. -1 for automatic (this is the default).");
 
 static int dsp_num_table[16];
 #endif
 
+#if defined(CONFIG_SND) || defined(CONFIG_SND_MODULE)
+char *alsa_id[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = NULL };
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
+MODULE_PARM(alsa_id, "1-" __MODULE_STRING(EM8300_MAX) "s");
+#else
+module_param_array(alsa_id, charp, NULL, 0444);
+#endif
+MODULE_PARM_DESC(alsa_id, "ID string for the audio part of the EM8330 chip (ALSA).");
+
+int alsa_index[EM8300_MAX] = { [ 0 ... EM8300_MAX-1 ] = -1 };
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
+MODULE_PARM(alsa_index, "1-" __MODULE_STRING(EM8300_MAX) "i");
+#else
+module_param_array(alsa_index, int, NULL, 0444);
+#endif
+MODULE_PARM_DESC(alsa_index, "Index value for the audio part of the EM8330 chip (ALSA).");
+#endif
+
 /* structure to keep track of the memory that has been allocated by
    the user via mmap() */
 struct memory_info
@@ -158,7 +261,11 @@ static struct pci_device_id em8300_ids[] = {
 
 MODULE_DEVICE_TABLE(pci, em8300_ids);
 
-static irqreturn_t em8300_irq(int irq, void *dev_id, struct pt_regs * regs)
+static irqreturn_t em8300_irq(int irq, void *dev_id
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
+			      , struct pt_regs * regs
+#endif
+			      )
 {
 	struct em8300_s *em = (struct em8300_s *) dev_id;
 	int irqstatus;
@@ -174,10 +281,14 @@ static irqreturn_t em8300_irq(int irq, void *dev_id, struct pt_regs * regs)
 
 		if (irqstatus & IRQSTATUS_VIDEO_FIFO) {
 			em8300_fifo_check(em->mvfifo);
+			em8300_video_interrupt(em);
 		}
 
 		if (irqstatus & IRQSTATUS_AUDIO_FIFO) {
-			em8300_fifo_check(em->mafifo);
+			em8300_audio_interrupt(em);
+			if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+			    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+				em8300_fifo_check(em->mafifo);
 		}
 
 		if (irqstatus & IRQSTATUS_VIDEO_VBL) {
@@ -190,6 +301,7 @@ static irqreturn_t em8300_irq(int irq, void *dev_id, struct pt_regs * regs)
 			em->tv = tv;
 			em->irqcount++;
 			wake_up(&em->vbi_wait);
+			em8300_vbl_interrupt(em);
 		}
 
 		write_ucregister(Q_IrqMask, em->irqmask);
@@ -218,7 +330,9 @@ static void release_em8300(struct em8300_s *em)
 	writel(0, &em->mem[0x2000]);
 
 	em8300_fifo_free(em->mvfifo);
-	em8300_fifo_free(em->mafifo);
+	if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+	    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+		em8300_fifo_free(em->mafifo);
 	em8300_fifo_free(em->spfifo);
 
 	/* free it */
@@ -237,7 +351,11 @@ static int em8300_io_ioctl(struct inode* inode, struct file* filp, unsigned int
 
 	switch (subdevice) {
 	case EM8300_SUBDEVICE_AUDIO:
-		return em8300_audio_ioctl(em, cmd, arg);
+		if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+		    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+			return em8300_audio_ioctl(em, cmd, arg);
+		else
+			return -EINVAL;
 	case EM8300_SUBDEVICE_VIDEO:
 		return em8300_video_ioctl(em, cmd, arg);
 	case EM8300_SUBDEVICE_SUBPICTURE:
@@ -276,14 +394,20 @@ static int em8300_io_open(struct inode* inode, struct file* filp)
 		em8300[card].nonblock[0] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
 		break;
 	case EM8300_SUBDEVICE_AUDIO:
-		em8300[card].nonblock[1] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
-		err = em8300_audio_open(em);
-		break;
+		if ((audio_driver_nr[card] == AUDIO_DRIVER_OSSLIKE)
+		    || (audio_driver_nr[card] == AUDIO_DRIVER_OSS)) {
+			em8300[card].nonblock[1] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
+			err = em8300_audio_open(em);
+			break;
+		}
+		else
+			return -ENODEV;
 	case EM8300_SUBDEVICE_VIDEO:
-		em8300[card].nonblock[2] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
+		em8300_require_ucode(em);
 		if (!em->ucodeloaded) {
 			return -ENODEV;
 		}
+		em8300[card].nonblock[2] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
 		em8300_video_open(em);
 
 		em8300_ioctl_enable_videoout(em, 1);
@@ -291,10 +415,11 @@ static int em8300_io_open(struct inode* inode, struct file* filp)
 		em8300_video_setplaymode(em, EM8300_PLAYMODE_PLAY);
 		break;
 	case EM8300_SUBDEVICE_SUBPICTURE:
-		em8300[card].nonblock[3] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
+		em8300_require_ucode(em);
 		if (!em->ucodeloaded) {
 			return -ENODEV;
 		}
+		em8300[card].nonblock[3] = ((filp->f_flags&O_NONBLOCK) == O_NONBLOCK);
 		err = em8300_spu_open(em);
 		break;
 	default:
@@ -327,9 +452,14 @@ static ssize_t em8300_io_write(struct file *file, const char * buf, size_t count
 		return em8300_video_write(em, buf, count, ppos);
 		break;
 	case EM8300_SUBDEVICE_AUDIO:
-		em->nonblock[1] = ((file->f_flags&O_NONBLOCK) == O_NONBLOCK);
-		return em8300_audio_write(em, buf, count, ppos);
-		break;
+		if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+		    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS)) {
+			em->nonblock[1] = ((file->f_flags&O_NONBLOCK) == O_NONBLOCK);
+			return em8300_audio_write(em, buf, count, ppos);
+			break;
+		}
+		else
+			return -ENODEV;
 	case EM8300_SUBDEVICE_SUBPICTURE:
 		em->nonblock[3] = ((file->f_flags&O_NONBLOCK) == O_NONBLOCK);
 		return em8300_spu_write(em, buf, count, ppos);
@@ -436,14 +566,17 @@ static unsigned int em8300_poll(struct file *file, struct poll_table_struct *wai
 
 	switch (subdevice) {
 	case EM8300_SUBDEVICE_AUDIO:
-		poll_wait(file, &em->mafifo->wait, wait);
-		if (file->f_mode & FMODE_WRITE) {
-			if (em8300_fifo_freeslots(em->mafifo)) {
-				pr_debug("Poll audio - Free slots: %d\n", em8300_fifo_freeslots(em->mafifo));
-				mask |= POLLOUT | POLLWRNORM;
+		if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+		    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS)) {
+			poll_wait(file, &em->mafifo->wait, wait);
+			if (file->f_mode & FMODE_WRITE) {
+				if (em8300_fifo_freeslots(em->mafifo)) {
+					pr_debug("Poll audio - Free slots: %d\n", em8300_fifo_freeslots(em->mafifo));
+					mask |= POLLOUT | POLLWRNORM;
+				}
 			}
+			break;
 		}
-		break;
 	case EM8300_SUBDEVICE_VIDEO:
 		poll_wait(file, &em->mvfifo->wait, wait);
 		if (file->f_mode & FMODE_WRITE) {
@@ -473,7 +606,9 @@ int em8300_io_release(struct inode* inode, struct file *filp)
 
 	switch (subdevice) {
 	case EM8300_SUBDEVICE_AUDIO:
-		em8300_audio_release(em);
+		if ((audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSSLIKE)
+		    || (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS))
+			em8300_audio_release(em);
 		break;
 	case EM8300_SUBDEVICE_VIDEO:
 		em8300_video_release(em);
@@ -516,6 +651,9 @@ struct file_operations em8300_fops = {
 	poll: em8300_poll,
 	open: em8300_io_open,
 	release: em8300_io_release,
+#if defined(CONFIG_EM8300_IOCTL32) && defined(HAVE_COMPAT_IOCTL)
+	compat_ioctl: em8300_compat_ioctl,
+#endif
 };
 
 #if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
@@ -652,7 +790,9 @@ static int init_em8300(struct em8300_s *em)
 		em->clockgen_overlaymode = CLOCKGEN_OVERLAYMODE_2;
 	}
 
-	pr_debug("em8300_main.o: activate_loopback: %d\n", activate_loopback);
+	em->clockgen = em->clockgen_tvmode;
+
+	pr_debug("em8300_main.o: activate_loopback: %d\n", activate_loopback[em->card_nr]);
 
 	return 0;
 }
@@ -711,14 +851,20 @@ static int __devinit em8300_probe(struct pci_dev *dev,
 	em8300_register_card(em);
 
 #if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
-	if ((em->dsp_num = register_sound_dsp(&em8300_dsp_audio_fops, dsp_num[em->card_nr])) < 0) {
-		printk(KERN_ERR "em8300: cannot register oss audio device!\n");
-	} else {
-		dsp_num_table[em->dsp_num >> 4 & 0x0f] = em8300_cards + 1;
-		pr_debug("em8300: registered dsp %i for device %i\n", em->dsp_num >> 4 & 0x0f, em8300_cards);
+	if (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS) {
+		if ((em->dsp_num = register_sound_dsp(&em8300_dsp_audio_fops, dsp_num[em->card_nr])) < 0) {
+			printk(KERN_ERR "em8300: cannot register oss audio device!\n");
+		} else {
+			dsp_num_table[em->dsp_num >> 4 & 0x0f] = em8300_cards + 1;
+			pr_debug("em8300: registered dsp %i for device %i\n", em->dsp_num >> 4 & 0x0f, em8300_cards);
+		}
 	}
 #endif
 
+#if defined(CONFIG_FW_LOADER) || defined(CONFIG_FW_LOADER_MODULE)
+	em8300_enable_card(em);
+#endif
+
 	em8300_cards++;
 	return 0;
 }
@@ -728,12 +874,18 @@ static void __devexit em8300_remove(struct pci_dev *pci)
 	struct em8300_s *em = pci_get_drvdata(pci);
 
 	if (em) {
+#if defined(CONFIG_FW_LOADER) || defined(CONFIG_FW_LOADER_MODULE)
+		em8300_disable_card(em);
+#else
 		if (em->ucodeloaded == 1)
 			em8300_disable_card(em);
+#endif
 
 #if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
-		unregister_sound_dsp(em->dsp_num);
+		if (audio_driver_nr[em->card_nr] == AUDIO_DRIVER_OSS)
+			unregister_sound_dsp(em->dsp_num);
 #endif
+
 		em8300_unregister_card(em);
 
 		release_em8300(em);
@@ -743,7 +895,7 @@ static void __devexit em8300_remove(struct pci_dev *pci)
 	pci_disable_device(pci);
 }
 
-static struct pci_driver em8300_driver = {
+struct pci_driver em8300_driver = {
 	.name     = "Sigma Designs EM8300",
 	.id_table = em8300_ids,
 	.probe    = em8300_probe,
@@ -752,10 +904,12 @@ static struct pci_driver em8300_driver = {
 
 static void __exit em8300_exit(void)
 {
-#ifdef CONFIG_EM8300_IOCTL32
+#if defined(CONFIG_EM8300_IOCTL32) && !defined(HAVE_COMPAT_IOCTL)
 	em8300_ioctl32_exit();
 #endif
 
+	em8300_preunregister_driver();
+
 	pci_unregister_driver(&em8300_driver);
 
 	unregister_chrdev(major, EM8300_LOGNAME);
@@ -766,6 +920,20 @@ static void __exit em8300_exit(void)
 static int __init em8300_init(void)
 {
 	int err;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
+	int i;
+	for (i=0; i < EM8300_MAX; i++)
+		if ((audio_driver[i]) && (audio_driver[i][0])) {
+			int j;
+			for (j=0; j < AUDIO_DRIVER_MAX; j++)
+				if (strcmp(audio_driver[i], audio_driver_name[j]) == 0) {
+					audio_driver_nr[i] = j;
+					break;
+				}
+			if (j == AUDIO_DRIVER_MAX)
+				printk(KERN_WARNING "em8300.o: Unknown audio driver: %s\n", audio_driver[i]);
+		}
+#endif /* ! CONFIG_MODULEPARAM */
 
 	//memset(&em8300, 0, sizeof(em8300) * EM8300_MAX);
 #if defined(CONFIG_SOUND) || defined(CONFIG_SOUND_MODULE)
@@ -800,7 +968,9 @@ static int __init em8300_init(void)
 		goto err_init;
 	}
 
-#ifdef CONFIG_EM8300_IOCTL32
+	em8300_postregister_driver();
+
+#if defined(CONFIG_EM8300_IOCTL32) && !defined(HAVE_COMPAT_IOCTL)
 	em8300_ioctl32_init();
 #endif
 
